/*
 * for special system call by special patch to QEMU
 */

#define SPECIAL_SYSCALL_READC  1
#define SPECIAL_SYSCALL_WRITEC 2
#define SPECIAL_SYSCALL_EXIT   (0x80 + 1)
#define SPECIAL_SYSCALL_OPEN   (0x80 + 3)
#define SPECIAL_SYSCALL_READ   (0x80 + 4)
#define SPECIAL_SYSCALL_WRITE  (0x80 + 5)
#define SPECIAL_SYSCALL_CLOSE  (0x80 + 6)

/*
 * See qemu/target/arm/translate-a64.c:disas_exc()
 */

#define HALT_GROUP  (10 << 25)
#define HALT_GROUP2 ( 6 << 29)
#define HALT_OPC2(opc2)   ((opc2)  <<  2)
#define HALT_LL(ll)       ((ll)    <<  0)
#define HALT_OPC(opc)     ((opc)   << 21)
#define HALT_IMM16(imm16) ((imm16) <<  5)

#define CALL_ExcpnGen(opc, opc2, ll, imm16) \
	.long (HALT_GROUP | HALT_GROUP2 | HALT_OPC2(opc2) | HALT_LL(ll) | \
	HALT_OPC(opc) | HALT_IMM16(imm16))

#define CALL_BRK      CALL_ExcpnGen(1, 0, 0, 0)
#define CALL_SEMIHOST CALL_ExcpnGen(2, 0, 0, 0xf000)
#define CALL_SWI(arg) CALL_ExcpnGen(0, 0, 1, (arg));

/*
 * for semihosting call on QEMU
 */

#define SEMIHOSTING_OPEN  0x01
#define SEMIHOSTING_CLOSE 0x02
#define SEMIHOSTING_WRITE 0x05
#define SEMIHOSTING_READ  0x06
#define SEMIHOSTING_EXIT  0x18

#define PUSH_REGS \
	sub	sp, sp, #32 ; \
	str	x3, [sp, #24] ; \
	str	x2, [sp, #16] ; \
	str	x1, [sp, #8] ; \
	str	x0, [sp]
#define POP_REGS \
	ldr	x3, [sp, #24] ; \
	ldr	x2, [sp, #16] ; \
	ldr	x1, [sp, #8] ; \
	add	sp, sp, #32

#if 0
	.org	0x800006f0
#else
	.org	0x000006f0
#endif

	.section .text

	.globl	_start
	.type	_start, %function
_start:
_fname:	.string	"flag.txt"
	.byte	0
	.byte	0
	.byte	0
	.long	0

	.quad	(1f + 0x80000000)	/* return address from main() */

1:	sub	sp, sp, #0x110	/* stack frame of main() */
	sub	sp, sp, #0x10	/* buffer in proc() */

	ldr	x0, _fname_addr
	orr	x0, x0, #0x80000000
	mov	x1, #(0xa - 1)
	add	x1, x1, #1	/* make 0xa, but avoid 0x0a('\n') */
	mov	x2, 0

	.globl	__semihosting_open
	.type	__semihosting_open, %function
__semihosting_open:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_OPEN
	CALL_SEMIHOST
	mov	x16, x0

	ldr	x1, _buffer_addr
	orr	x1, x1, #0x80000000
	mov	w0, #('\n' - 1)
	add	w0, w0, #1	/* make '\n', but avoid 0x0a('\n') */
	strb	w0, [x1, #31]

	mov	x0, x16
	mov	x2, 32

	.globl	__semihosting_write
	.type	__semihosting_write, %function
__semihosting_write:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_WRITE
	CALL_SEMIHOST

	mov	x0, x16

	.globl	__semihosting_close
	.type	__semihosting_close, %function
__semihosting_close:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_CLOSE
	CALL_SEMIHOST

	mov	x0, #0

	.globl	__semihosting_exit
	.type	__semihosting_exit, %function
__semihosting_exit:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_EXIT
	CALL_SEMIHOST

1:	b	1b

	.align 8
_fname_addr:
	.quad	_fname
_buffer_addr:
	.quad	_buffer
_buffer:
	.string	"AArch64 sample 56789012345678901"
	.align 8
