/*
 * for special system call by special patch to QEMU
 */

#define SPECIAL_SYSCALL_READC  1
#define SPECIAL_SYSCALL_WRITEC 2
#define SPECIAL_SYSCALL_EXIT   (0x80 + 1)
#define SPECIAL_SYSCALL_OPEN   (0x80 + 3)
#define SPECIAL_SYSCALL_READ   (0x80 + 4)
#define SPECIAL_SYSCALL_WRITE  (0x80 + 5)
#define SPECIAL_SYSCALL_CLOSE  (0x80 + 6)

/*
 * See qemu/target/arm/translate-a64.c:disas_exc()
 */

#define HALT_GROUP  (10 << 25)
#define HALT_GROUP2 ( 6 << 29)
#define HALT_OPC2(opc2)   ((opc2)  <<  2)
#define HALT_LL(ll)       ((ll)    <<  0)
#define HALT_OPC(opc)     ((opc)   << 21)
#define HALT_IMM16(imm16) ((imm16) <<  5)

#define CALL_ExcpnGen(opc, opc2, ll, imm16) \
	.long (HALT_GROUP | HALT_GROUP2 | HALT_OPC2(opc2) | HALT_LL(ll) | \
	HALT_OPC(opc) | HALT_IMM16(imm16))

#define CALL_BRK      CALL_ExcpnGen(1, 0, 0, 0)
#define CALL_SEMIHOST CALL_ExcpnGen(2, 0, 0, 0xf000)
#define CALL_SWI(arg) CALL_ExcpnGen(0, 0, 1, (arg));

	.section .text

	.globl	_start
	.type	_start, %function
_start:
	ldr	x0, _stack_addr
	mov	sp, x0
	bl	main

	/*
	 * Registers as return value and first argument are same.
	 * Setup of first argument is unneed.
	 */
	/* fall through */

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__exit
	.type	__exit, %function
__exit:
	CALL_SWI(SPECIAL_SYSCALL_EXIT)
	ret
#else
	bl	__semihosting_exit
1:	b	1b
#endif

	.globl	__readc
	.type	__readc, %function
__readc:
	CALL_SWI(SPECIAL_SYSCALL_READC)
	ret

	.globl	__writec
	.type	__writec, %function
__writec:
	CALL_SWI(SPECIAL_SYSCALL_WRITEC)
	ret

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__read
	.type	__read, %function
__read:
	CALL_SWI(SPECIAL_SYSCALL_READ)
	ret

	.globl	__write
	.type	__write, %function
__write:
	CALL_SWI(SPECIAL_SYSCALL_WRITE)
	ret

	.globl	__open
	.type	__open, %function
__open:
	CALL_SWI(SPECIAL_SYSCALL_OPEN)
	ret

	.globl	__close
	.type	__close, %function
__close:
	CALL_SWI(SPECIAL_SYSCALL_CLOSE)
	ret
#endif

/*
 * for semihosting call on QEMU
 */

#define SEMIHOSTING_OPEN  0x01
#define SEMIHOSTING_CLOSE 0x02
#define SEMIHOSTING_WRITE 0x05
#define SEMIHOSTING_READ  0x06
#define SEMIHOSTING_EXIT  0x18

#define PUSH_REGS \
	sub	sp, sp, #32 ; \
	str	x3, [sp, #24] ; \
	str	x2, [sp, #16] ; \
	str	x1, [sp, #8] ; \
	str	x0, [sp]
#define POP_REGS \
	ldr	x3, [sp, #24] ; \
	ldr	x2, [sp, #16] ; \
	ldr	x1, [sp, #8] ; \
	add	sp, sp, #32

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__semihosting_open
	.type	__semihosting_open, %function
__semihosting_open:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_OPEN
	CALL_SEMIHOST
	POP_REGS
	ret

	.globl	__semihosting_close
	.type	__semihosting_close, %function
__semihosting_close:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_CLOSE
	CALL_SEMIHOST
	POP_REGS
	ret

	.globl	__semihosting_write
	.type	__semihosting_write, %function
__semihosting_write:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_WRITE
	CALL_SEMIHOST
	POP_REGS
	ret

	.globl	__semihosting_read
	.type	__semihosting_read, %function
__semihosting_read:
	PUSH_REGS
	mov	x1, sp
	mov	w0, #SEMIHOSTING_READ
	CALL_SEMIHOST
	POP_REGS
	ret
#endif

	.globl	__semihosting_exit
	.type	__semihosting_exit, %function
__semihosting_exit:
#ifdef SYSCALL_WRAPPER_ALL
	PUSH_REGS
	mov	x1, sp
#else
	ldr	x1, _zero_addr
#endif
	mov	w0, #SEMIHOSTING_EXIT
	CALL_SEMIHOST
#ifdef SYSCALL_WRAPPER_ALL
	POP_REGS
#endif
	ret

	.align 8
_zero:
	.quad	0
_zero_addr:
	.quad	_zero
_stack_addr:
	.quad	_estack
