/*
 * for special system call by special patch to QEMU
 */

#define SPECIAL_SYSCALL_READC  1
#define SPECIAL_SYSCALL_WRITEC 2
#define SPECIAL_SYSCALL_EXIT   (0x80 + 1)
#define SPECIAL_SYSCALL_OPEN   (0x80 + 3)
#define SPECIAL_SYSCALL_READ   (0x80 + 4)
#define SPECIAL_SYSCALL_WRITE  (0x80 + 5)
#define SPECIAL_SYSCALL_CLOSE  (0x80 + 6)

	.section .text

	.globl	_start
	.type	_start, %function
_start:
	ldr	sp, _stack_addr
	bl	main

	/*
	 * Registers as return value and first argument are same.
	 * Setup of first argument is unneed.
	 */
	/* fall through */

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__exit
	.type	__exit, %function
__exit:
	swi	SPECIAL_SYSCALL_EXIT
	mov	pc, lr
#else
	bl	__semihosting_exit
1:	b	1b
#endif

	.globl	__readc
	.type	__readc, %function
__readc:
	swi	SPECIAL_SYSCALL_READC
	mov	pc, lr

	.globl	__writec
	.type	__writec, %function
__writec:
	swi	SPECIAL_SYSCALL_WRITEC
	mov	pc, lr

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__read
	.type	__read, %function
__read:
	swi	SPECIAL_SYSCALL_READ
	mov	pc, lr

	.globl	__write
	.type	__write, %function
__write:
	swi	SPECIAL_SYSCALL_WRITE
	mov	pc, lr

	.globl	__open
	.type	__open, %function
__open:
	swi	SPECIAL_SYSCALL_OPEN
	mov	pc, lr

	.globl	__close
	.type	__close, %function
__close:
	swi	SPECIAL_SYSCALL_CLOSE
	mov	pc, lr
#endif

/*
 * for semihosting call on QEMU
 */

#define SWI_SEMIHOSTING  0x123456
#define SEMIHOSTING_OPEN  0x01
#define SEMIHOSTING_CLOSE 0x02
#define SEMIHOSTING_WRITE 0x05
#define SEMIHOSTING_READ  0x06
#define SEMIHOSTING_EXIT  0x18

#ifdef SYSCALL_WRAPPER_ALL
	.globl	__semihosting_open
	.type	__semihosting_open, %function
__semihosting_open:
	push	{ r0, r1, r2, r3 }
	mov	r1, sp
	ldr	r0, _semihosting_sys_open
	swi	SWI_SEMIHOSTING
	pop	{ r1 }
	pop	{ r1, r2, r3 }
	mov	pc, lr

	.globl	__semihosting_close
	.type	__semihosting_close, %function
__semihosting_close:
	push	{ r0, r1, r2, r3 }
	mov	r1, sp
	ldr	r0, _semihosting_sys_close
	swi	SWI_SEMIHOSTING
	pop	{ r1 }
	pop	{ r1, r2, r3 }
	mov	pc, lr

	.globl	__semihosting_write
	.type	__semihosting_write, %function
__semihosting_write:
	push	{ r0, r1, r2, r3 }
	mov	r1, sp
	ldr	r0, _semihosting_sys_write
	swi	SWI_SEMIHOSTING
	pop	{ r1 }
	pop	{ r1, r2, r3 }
	mov	pc, lr

	.globl	__semihosting_read
	.type	__semihosting_read, %function
__semihosting_read:
	push	{ r0, r1, r2, r3 }
	mov	r1, sp
	ldr	r0, _semihosting_sys_read
	swi	SWI_SEMIHOSTING
	pop	{ r1 }
	pop	{ r1, r2, r3 }
	mov	pc, lr
#endif

	.globl	__semihosting_exit
	.type	__semihosting_exit, %function
__semihosting_exit:
#ifdef SYSCALL_WRAPPER_ALL
	push	{ r0, r1, r2, r3 }
	mov	r1, sp
#else
	ldr	r1, _zero_addr
#endif
	ldr	r0, _semihosting_sys_exit
	swi	SWI_SEMIHOSTING
#ifdef SYSCALL_WRAPPER_ALL
	pop	{ r1 }
	pop	{ r1, r2, r3 }
#endif
	mov	pc, lr

	.align 4
#ifdef SYSCALL_WRAPPER_ALL
_semihosting_sys_open:
	.long	SEMIHOSTING_OPEN
_semihosting_sys_close:
	.long	SEMIHOSTING_CLOSE
_semihosting_sys_write:
	.long	SEMIHOSTING_WRITE
_semihosting_sys_read:
	.long	SEMIHOSTING_READ
#endif
_semihosting_sys_exit:
	.long	SEMIHOSTING_EXIT

	.align 4
_zero:
	.long	0
_zero_addr:
	.long	_zero
_stack_addr:
	.long	_estack
