#include "syscall.h"

long data_value = 0xabadface;
int bss_value;

#ifdef USE_SYSCALL_WRAPPER
void __exit(int status)
{
  __r_exit(0, status);
}
int __read(int fd, void *buffer, int size)
{
  return __r_read(0, fd, buffer, size);
}
int __write(int fd, const void *buffer, int size)
{
  return __r_write(0, fd, buffer, size);
}
int __open(const char *path, int flags, int mode)
{
  return __r_open(0, path, flags, mode);
}
int __close(int fd)
{
  return __r_close(0, fd);
}
#endif

void exit(int status)
{
  __exit(status);
}

int write1(int fd, unsigned char c)
{
  return __write(fd, &c, 1);
}

int putchar(int fd, int c)
{
  write1(fd, c);
  return c;
}

int puts(int fd, char *str)
{
  for (; *str; str++)
    putchar(fd, *str);
  return 0;
}

#ifdef RSHIFT
/*
 * For architecture not to be able to use shift instruction for well,
 * execute shift operation without shift instruction.
 */
unsigned long rshift1(unsigned long value)
{
  unsigned long mask, bit = 1;
  unsigned long ret = 0;
  int i;

  for (i = 0; i < sizeof(unsigned long) * 8 - 1; i++) {
    mask = bit + bit; /* Use addition for no using shift and multiplication */
    if (value & mask)
      ret |= bit;
    bit = bit + bit;
  }

  return ret;
}

unsigned long RSHIFT(unsigned long value)
{
  int num = 4;
  for (; num > 0; num--)
    value = rshift1(value);
  return value;
}
#endif

int putxval(int fd, unsigned long value, int column)
{
  char buf[17];
  char *p;

  p = buf + sizeof(buf) - 1;
  *(p--) = '\0';

  if (!value && !column)
    column++;

  while (value || column) {
    *(p--) = "0123456789abcdef"[value & 0xf];
    value >>= 4;
    if (column) column--;
  }

  puts(fd, p + 1);

  return 0;
}

int __readc();
int __writec(int c);
void __semihosting_exit();
int __semihosting_read(int fd, void *buffer, int size);
int __semihosting_write(int fd, const void *buffer, int size);
int __semihosting_open(const char *path, int flags, int mode);
int __semihosting_close(int fd);

int main()
{
  int fd = 1, r;
  char buf[32];

  puts(fd, "Hello World! ");
  putxval(fd, data_value, 0);
  puts(fd, " This architecture is " ARCH "\n");

  puts(1, "******** special system call test ********\n");
  __writec('1'); __writec('c'); __writec('h'); __writec('a'); __writec('r');
  __writec('\n');

  __write(1, "write test\n", 11);

  puts(1, "******** semihosting test ********\n");

  puts(1, "**** standard in/out test ****\n");
  __semihosting_write(1, "Input and push ENTER:\n", 22);
  r = __semihosting_read(0, buf, sizeof(buf));
#if defined(__arm__) || defined(__aarch64__)
  r = (r < 0) ? r : sizeof(buf) - r;
#endif
  putxval(1, r, 0); puts(1, "\n");
  __semihosting_write(1, buf, r);

  puts(1, "**** file write test ****\n");
#if defined(__arm__) || defined(__aarch64__)
  fd = __semihosting_open("sample.txt", 6, 0);
#endif
#ifdef __mips__
  fd = __semihosting_open("sample.txt", 0x602, 0);
#endif
  __semihosting_write(fd, "file test\n", 10);
  __semihosting_close(fd);

  puts(1, "**** file read test ****\n");
  fd = __semihosting_open("sample.txt", 0, 0);
  r = __semihosting_read(fd, buf, sizeof(buf));
#if defined(__arm__) || defined(__aarch64__)
  r = (r < 0) ? r : sizeof(buf) - r;
#endif
  putxval(1, r, 0); puts(1, "\n");
  __semihosting_write(1, buf, r);
  __semihosting_close(fd);

  puts(1, "**** exit test ****\n");
  __semihosting_exit();
  puts(1, "error\n");

  exit(0);
  return 0;
}
